/*
   $Id: Bank.cpp 672 2011-12-09 13:53:53Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "utils.h"
#include "Bank.h"
#include "IndexTree.h"
#include "BankBuffer.h"

#define INDEX_VERSION "1.0"

/**
 * \brief The default constructor asks __parameters for the name of the index, and if the bank will have to be indexed
 *        
 */

Bank::Bank():
	__parameters(Parameters::Instance()),
	__mask(Mask::Instance()),
	__bank_name(__parameters.GetBankName()),
	__filename(__parameters.GetBankFilename()),
	Words(0),
	Length(0),
	Positions(0),
	nb_seq(0),
	seq_info(),
	complexity(__parameters.GetComplexity()),
	bi_bfr(NULL),
	blist_aptr(NULL),
	tree_ptr(NULL)
{
	// Indexation mode: in indexation mode, we have some data structures to initialize
	// __SetNbPass is called to set the number of indexation passes, depend on the bank file length
	if ( __parameters.IsIndexMode() )
	{
		__init_node_width();
		__mask.Initialize(__parameters.GetMskStr());
		__SetNbPass();
	}
	
	// Read index file mode: in this mode, we read the .he file to init some fields
	else
	{
		string str_msk;
		IstreamPtr is_ptr(__filename.GetName(),HE_EXTENSION,__filename.GetCompression());
		istream& is = *is_ptr;
	    uint64_t nb_read_sequences=0;
	    while (is)
	    {
			string field_name;
			is >> field_name;
			if ( field_name == "VERSION")
			{
				float version_nb;
				is >> version_nb;
				if ( version_nb != 1.0 )
				{
					throw ( runtime_error( " ERROR - BAD VERSION NUMBER"));
				}
			}
			else if ( field_name == "NAME" ) {
				is >> __bank_name;
			}
			else if ( field_name == "MASK" ) {
				is >> str_msk;
				__mask.Initialize(str_msk);
			}
			else if ( field_name == "INDEX_SIZE" ) {
				is >> Size;
			}
			else if ( field_name == "WORDS" ) {
				is >> Words;
			}
			else if ( field_name == "LENGTH" ) {
				is >> Length;
			}
			else if ( field_name == "SEQ" ) {
				is >> nb_seq;
			}
			else if ( field_name == "COMPLEXITY" ) {
				is >> complexity;
			}
			else if ( field_name == "NODE_WIDTH" ) {
				for (int i=ROOT_LEVEL;i<=MAX_LEAF_LEVEL;i++)
				is >> this->node_width[i];
			}
			else if (field_name == "SEQUENCE") {
				SequencePosition sp;
				is >> sp;
				seq_info.push_back(sp);
				++nb_read_sequences;
			}
		}
		if (nb_read_sequences != nb_seq)
		{
			string msg = "ERROR - number of declared sequences different from number of read sequences in the file ";
			msg += __filename.GetName() + "." + HE_EXTENSION;
			throw(runtime_error( msg ));
		}
			
		// If necessary, init datastructure for reading the black list
		if 	(blist_aptr.get() == NULL)
		{
			blist_aptr= auto_ptr<BlackList> (new BlackList(BLACK_LIST_SIZE));
			// Read the black-listed word file
			IstreamPtr blck_ptr(__filename.GetName(), BL_EXTENSION, __filename.GetCompression());
			*blck_ptr >> *blist_aptr;
		}			
	}
	
	// The old releases of glintindex did not store the end position of the , thus the SAM header will be false
	if ( seq_info.IsOldIndex() && __parameters.GetOutputFormat() == OUTPUT_FORMAT_SAM )
	{
		throw runtime_error( "ERROR - You CANNOT USE the format SAM with an Old index. Please run glintindex and try again" );
	}
}
		
/**
 * \brief Given a position in the Bank (.fa file), return the corresponding SequencePosition
*/

SequencePosition Bank::GetTrueSeqPos(uint32_t pos) const
{
	vector<SequencePosition>::const_iterator left,right,middle;
	uint64_t size=GetNbSeq();
	int64_t spos = pos; // Avoiding a warning

	left=seq_info.begin();
	right=left+size-1;
	
	while (size>1) {
		middle=left+(uint64_t)(size/2);
		if (spos < middle->start) {
			right = middle;
		} else {
			left = middle;
		}
		size=(uint64_t)(size+1)/2;
	}
	
	return *left;
}

/**
 * \brief Bank output operator
 * \param [out] os An output stream
 * \param [in]  db a Bank to print
 */

ostream & operator<<(ostream& os, Bank const & db) {
	os << "VERSION " << INDEX_VERSION << '\n';
    os << "NAME " << db.__bank_name << '\n';
    os << "MASK "<< db.__mask.GetString() << '\n';
    os << "INDEX_SIZE "<< db.Size << '\n';
    os << "WORDS "<< db.Words << '\n';
    os << "LENGTH " << db.Length << '\n';
    os << "SEQ " << db.nb_seq <<'\n';
    os << "COMPLEXITY " << db.complexity << '\n';
    os << "NODE_WIDTH ";
    for (int i=ROOT_LEVEL;i<=MAX_LEAF_LEVEL;i++)
        os << db.node_width[i] << ' ';
    os << '\n';
    for (unsigned int i=0; i < db.nb_seq; i++)
        os << "SEQUENCE " << db.seq_info[i] << '\n';
    return os;
}

/**
 * \brief Set the sequence length
 * \pre   The indexation process must NOT be started ! (this should be checked, but is NOT)
 * \param The buffer length (bytes, taking into account the separators)
 * \param The sequence length (nb of nucletoides)
 */
void Bank::SetSequenceLength(uint32_t bfr_length, uint32_t seq_length) 
{ 
	Length           = bfr_length;
	SequenceLength   = seq_length;
}
//void Bank::SetSequenceLength(uint32_t bfr_length, uint32_t seq_length) 
//{ 
	//Length           = bfr_length;
	//SequenceLength   = seq_length;
	//uint16_t pass_nb = __parameters.GetIndexNbPass();
	//bool silent      = __parameters.IsSilent();
	
	//// Default value: we must compute a correct value
	//if (pass_nb == 0)
	//{
		//if ( seq_length < 1000000000 ) // if < 1 Go, 1 pass is enough
		//{
			//pass_nb = 1;
		//}
		//else if ( seq_length < 2000000000 ) // if < 2 Go, 2 pass is enough
		//{
			//pass_nb = 2;
		//}
		//else
		//{
			//pass_nb = 4;
		//}
		//if ( !silent)
		//{
			//cerr << "Computing the pass number: " << pass_nb << "\n";
		//}		
	//}
	//tree_ptr -> SetNbPass(pass_nb);
//}


/**
 * \brief Set the number of passes for the indexation
 * \pre   The indexation process must NOT be started ! (this should be checked, but is NOT)
 */
void Bank::__SetNbPass()
{
	uint16_t nb_pass = __parameters.GetIndexNbPass();
	bool silent      = __parameters.IsSilent();
	CompressionAlgorithm compression = __filename.GetCompression();
	string file_name = __filename.GetName();
	if ( compression != NOCOMPRESSION )
	{
		file_name += '.';
		file_name += __filename.GetCompressionExtension();
	}

	if ( nb_pass == 0)	
	{
		off_t size = 0;
		struct stat buf;
		if (!stat ( file_name.c_str(), &buf ))
		{
			size = buf.st_size;
			if ( compression != NOCOMPRESSION ) 
			{
				size *= 4;
			}
			if ( size < 1000000000 ) // if < 1 Go, 1 pass is enough
			{
				nb_pass = 1;
			}
			else if ( size < 2000000000 ) // if < 2 Go, 2 pass is enough
			{
				nb_pass = 2;
			}
			else
			{
				nb_pass = 4;
			}
			if ( !silent)
			{
				cerr << "Computing the pass number: " << nb_pass << "\n";
			}				
		}
		else
		{
			string msg = "ERROR - Cannot know the size of file ";
			msg += file_name;
			throw ios_base::failure(msg);
		}
	}
	else
	{
		if ( !silent )
		{
			cerr << "pass number read from the -p switch: " << nb_pass << "\n";
		}
	}
	__nb_pass = nb_pass;
}

/**
 * \brief The general indexation method
 *        - Index the data, using the IndexSequence() method from the BankBuffer object
 *        - Unless otherwise specified, write the indexes to some files
 */ 
void Bank::Index(bool memory)
{
	// Refuse indexation if not in indexation mode
	if ( ! __parameters.IsIndexMode() )
	{
		throw(logic_error("ERROR - The Bank object is not initialized for indexation"));
	}

	// If necessary, init datastructures useful for indexation
	if 	(blist_aptr.get() == NULL)
	{
		blist_aptr= auto_ptr<BlackList> (new BlackList(BLACK_LIST_SIZE));
		tree_ptr = IndexTree::InstancePtr();
	}
	
	// Indexing data
	BankBuffer bfr(*blist_aptr,memory);
	bfr.IndexSequence();

	// If memory is true, nothing is stored on disk	
    if ( !memory)
    {
	    /* Writing the Bank parameters to the .he file */
	    OstreamPtr out_ptr (__filename.GetName(), HE_EXTENSION, __filename.GetCompression());
	    *out_ptr << *this;
    
		/* Writing the deleted words inside the black list */
		OstreamPtr blck_ptr(__filename.GetName(), BL_EXTENSION, __filename.GetCompression());
		*blck_ptr << *blist_aptr;		
	}
			
	// FOR DEBUG ONLY !!!
	// cout << tree;
}
