/*
   $Id: CompactTree.cpp 680 2012-01-30 14:42:53Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

using namespace std;
#include <iostream>
#include <cstdio>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fstream>
#include <iomanip>
#include <cerrno>
#include <map>

#ifdef _OPENMP
   #include <omp.h>
#endif

#include "ptypes.h"
#include "utils.h"
#include "CompactTree.h"
const int8_t SStrand[2]={1,-1};

extern RealLogger* logger;

/*
read_patch
Usage:    same as the read function (see man 2 read)
Function: Cut a read operation in several slices if count is too high
          I think there is a bug in read on some systems, at least for linux ia64

*/

ssize_t read_patch(int fd, const void *buf, size_t count)
{
	Parameters & parameters = Parameters::Instance();
	
	size_t max_count = 100000000;
	ssize_t rvl=0;
	while (count>0)
	{
		ssize_t r;
		size_t c;
		char * b;
		c = count>=max_count ? max_count : count;
		r = read(fd,(void*)buf,c);
		if (r==-1)
		{
			break;
		}
		else
		{
			rvl += r;
		};
		if (count<max_count)
		{
			break;
		}
		else
		{
			count -= max_count;
		};
		b = (char*)buf + max_count;
		buf = (void*) b;
		if ( ! parameters.IsSilent() )
		{
			cerr << rvl << "\r";
		}
	};
	return rvl;
}

/* From this point to the end of this file, all read calls will be replaced with read_patch */

#define read read_patch


/*=====================================================================
							chargement de l'index
=======================================================================*/

CompactTree::CompactTree():
	_parameters(Parameters::Instance()),
	_bank(Bank::Instance()),
	_mask(Mask::Instance()),
	//2**n
	nb_top(1<<(2*_mask.GetLength(ROOT_LEVEL))),InitedFromBuffer(false) // const vire provisoirement

{
    /*
        We check to know if _bank maintains buffers for the binary index and for the data
    */

    if (_bank.GetBiBfr() != NULL)  // Yes... Use those buffers
        __InitFromBfr();
    else
        __InitFromFiles();      // No... Read the index files

    /*
        Init the __Ptr array of functions, using the info found in _bank.node_width
    */

    __Ptr[0] = NULL;     // Not used
    //__Ptr[ROOT_LEVEL] = (_bank.GetNodeWidth(ROOT_LEVEL) == 5) ? __Ptr4 : __Ptr2;  // nb_son(1b + ptr (4b or 2b)

    //cout<<"Root level width  : "<<_bank.GetNodeWidth(ROOT_LEVEL)<<"\n";

    switch(_bank.GetNodeWidth(ROOT_LEVEL)) {
    	case 9 : __Ptr[ROOT_LEVEL] = __Ptr8;break;
		case 5 : __Ptr[ROOT_LEVEL] = __Ptr4;break;
		case 3 : __Ptr[ROOT_LEVEL] = __Ptr2;break;
	}

    for (int i=ROOT_LEVEL+1; i<MAX_LEAF_LEVEL+1; i++)
    {
        switch(_bank.GetNodeWidth(i))
        {
            case 6: __Ptr[i] = __Ptr4; break; // node = hash_code (1b) + nb_son(1b) + ptr(4b)
            case 4: __Ptr[i] = __Ptr2; break; // node = hash_code (1b) + nb_son(1b) + ptr(2b)
        };
    }
    
    
}

/**
 * \brief Init CompactTree from a buffer, useful in auto_indexation mode. Called by the constructor
*/

void CompactTree::__InitFromBfr()
{
    base_index = (unsigned char*) _bank.GetBiBfr()->GetBase();
    index=base_index + _bank.GetSize() - _bank.GetNodeWidth(ROOT_LEVEL);   /*    First go to the end !!   */
    base_seq   = (unsigned char*) _bank.GetFaBfr()->GetBase();
   	seq_length = _bank.GetLength();
    InitedFromBuffer=true;
}

/**
 * \brief Init CompactTree from the index files, useful if NOT in auto_indexation mode. Called by the constructor
*/

void CompactTree::__InitFromFiles()
{
    // Read the .bi file
	string index_name                = _bank.GetFileName().GetName();
	CompressionAlgorithm compression = _bank.GetFileName().GetCompression();

	FILE* fh_file     = FOpen ( index_name, BI_EXTENSION, "r", compression);
	base_index        = new unsigned char[_bank.GetSize()];
	ssize_t lus       = fread( (void*) base_index, sizeof(char), _bank.GetSize(), fh_file);
	fclose ( fh_file);

	if (lus==-1)
	{
		string msg = (string) "ERROR reading file " + index_name + "." + BI_EXTENSION;
		throw(runtime_error(msg));
	};
	if (lus != _bank.GetSize())
	{
		ostringstream err;
		err << (string )"ERROR reading file " + index_name + "." + BI_EXTENSION << " read " << lus << " should be " << _bank.GetSize() << '\n';
		throw(runtime_error(err.str()));
	}; 
	if (!_parameters.IsSilent())
		cerr<<lus<<" read and "<<_bank.GetSize()<<" to read \n";

	index_size = _bank.GetSize()+1;
	index      = base_index + _bank.GetSize() - _bank.GetNodeWidth(ROOT_LEVEL);   /*    First go to the end !!   */

    // Read the .fa file
    fh_file  = FOpen ( index_name, FA_EXTENSION, "r", compression);
    base_seq = new unsigned char[_bank.GetLength()+1];
	lus      = fread( (void*) base_seq, sizeof(char), _bank.GetLength(), fh_file);
	fclose(fh_file);
    
    if (!_parameters.IsSilent())
		cerr<<lus<<" read and "<<_bank.GetLength()<<" to read \n";

	if (lus==-1)
	{
        string msg = (string) "ERROR in reading file " + index_name + "." + FA_EXTENSION;
        throw(runtime_error(msg));
    }
	seq_length = _bank.GetLength();
}

/*
    The destructor
*/

CompactTree::~CompactTree()
{
	if (InitedFromBuffer==false)
	{
		delete[] base_index;
		delete[] base_seq;
	}
}



/*=====================================================================
							recherche  dans l'arbre compact
=======================================================================*/


/**
 * \brief A Seed is a word found in the index tree
 *            - The match can be kept at different levels in the tree.
 *            - The deeper the level, the more specific the match
 */
typedef const unsigned char* Seed;

/**
 * \brief TreePath is used to temporarily keep the whole path of the match, and compute the hit of the match

*/
/*
struct TreePath {
    void Reset(short s) {   // Resetting the seeds
        for (int i=0;i<MAX_LEAF_LEVEL+1;i++)
            ancestor_nodes[i].Reset(s);
//        Deepest.Reset(s);
    };
	uint8_t Hit;
	Seed ancestor_nodes[MAX_LEAF_LEVEL+1];
//	Seed Deepest;
};
*/
class TreePath
{
	public:
		//TreePath():__hit(0) { __ancestor_nodes.reserve(MAX_LEAF_LEVEL); for (size_t i=0; i<ROOT_LEVEL;i++) __ancestor_nodes.push_back(NULL); };
		//void Push(Seed s) { __ancestor_nodes.push_back(s); };

		TreePath():__hit(0),__level(ROOT_LEVEL-1) { __ancestor_nodes.reserve(MAX_LEAF_LEVEL); for (size_t i=0; i<MAX_LEAF_LEVEL;i++) __ancestor_nodes.push_back(NULL); };
		void Push(Seed s) { __ancestor_nodes[++__level] = s; };
		

		uint8_t GetHit() const { return __hit; };
		void  FixHit(Seed p, uint8_t l) { if(p==__ancestor_nodes[l]) __hit=l; };
		Seed operator[](size_t l) const { return __ancestor_nodes[l]; };
	private:
		uint8_t __hit;
		uint8_t __level;
		vector<Seed> __ancestor_nodes;
};
//typedef TreePath vector<Seed1>;


/** \brief Compare the window with the compressed index
 * \param w The window encapsulating the word to compare
 * \param qpos The position of this word in the query
 * \param glintstrandengine Storing the positions inside a strand engine
 * \param query The QueryBuffer
 * \public
*/

void CompactTree::CompareWindow(const Window &w,uint32_t qpos,GlintEngineSingleStrand& strand_engine, const QueryBuffer& query) const
{
	int bstrand = (strand_engine.GetStrand()!=PLUS); // PLUS => 0, MINUS => 1
	//tree_var.Reset(SStrand[bstrand]);

	TreePath tree_path;
	//for (size_t i=0; i<ROOT_LEVEL;i++) tree_path.push_back(NULL);
	
	int hit = (bstrand==0)?__Search(w, tree_path):__ReverseSearch(w, tree_path);
	//int min = _mask.GetLeafLevel();
	NbPos=0;                        // Necessary for GlintFreq

	if ( hit >= _parameters.GetLowerLevel() )
	{	
		// TODO Remove : replace DeepestLevel and Lowerlevel by a single level
		int idx=( hit < _parameters.GetDeepestLevel() ) ? hit : _parameters.GetDeepestLevel();
		// tree_var.Deepest=tree_var.ancestor_nodes[idx];
		//min=(idx<min) ? idx : min;
		
		//__RetrievePos((const unsigned char*)tree_var.ancestor_nodes[min].node,min,qpos,strand_engine,query);
		__RetrievePos(tree_path[idx],idx,tree_path,qpos,strand_engine,query);
	}
}

/**
 * \brief Search a word in the compact tree, strand \b +, starting at the root level
 *        Called by CompareWindow
 * \param [in] window The window encapsulating the word to search
 * \param [inout] tree_path The path of the search in the tree
 * \return The level at which the match is found, 0 if no match found
 * \private
*/

int CompactTree::__Search(const Window& window, TreePath& tree_path ) const
{
	uint32_t code=window.TreeHash(ROOT_LEVEL);
    const unsigned char *p = GetTop(code);
    uint8_t sons= GetTopNbSons(p);
    uint8_t hit = 0;

    bool found = false;
	if (sons) {
		//tree_var.ancestor_nodes[ROOT_LEVEL].level=ROOT_LEVEL;
		//tree_var.ancestor_nodes[ROOT_LEVEL].node=p;
		tree_path.Push(p);
		
		found=true;
#ifdef DEBUGLOG
		if (found)
		{
			logger->SetLogType("CV") << "CompactTree::__Search: Word found "<<window.GetWord().Convert()<< " Level: " << (int) hit << "\n";
		}
#endif
		hit = __Search(window,GetTopSons(p),sons,ROOT_LEVEL+1,tree_path);
	}
	hit=(hit>ROOT_LEVEL) ? hit : ROOT_LEVEL;

	return (found) ? hit : 0;
}

/**
 * \brief Search a word in the compact tree, strand \b -, starting at the root level
 *        Called by CompareWindow
 * \param [in] window The window encapsulating the word to search
 * \param [inout] tree_path The path of the search in the tree
 * \return The level at which the match is found, 0 if no match found
 * \private
*/

int CompactTree::__ReverseSearch(const Window& window, TreePath& tree_path) const
{
	uint32_t code=window.ReverseTreeHash(ROOT_LEVEL);
    const unsigned char *p = GetTop(code);
    uint8_t sons= GetTopNbSons(p);
    uint8_t hit = 0;

    bool found = false;
	if (sons) {
		//tree_var.ancestor_nodes[ROOT_LEVEL].level=ROOT_LEVEL;
		//tree_var.ancestor_nodes[ROOT_LEVEL].node=p;
		tree_path.Push(p);
		
		found=true;
		hit = __ReverseSearch(window,GetTopSons(p),sons,ROOT_LEVEL+1,tree_path);
	}
	hit=(hit>ROOT_LEVEL) ? hit : ROOT_LEVEL;
#ifdef DEBUGLOG
		if (found)
		{
			logger->SetLogType("CV") << "CompactTree::__ReverseSearch: Word found "<<window.GetWord().Convert()<< " Level: " << (int) hit << "\n";
		}
#endif

	return (found) ? hit : 0;
}

/**
 * \brief Search a word in the compact tree, strand \b +, starting at any level higher than root level
 *        Called by __Search (1st version)
 * \param [in] window The window encapsulating the word to search
 * \param [in] p The starting node
 * \param [in] The number of brothers (including p)
 * \param [in] level The level in the tree (at least ROOT_LEVEL+1)
 * \param [inout] tree_path The path of the search in the tree
 * \return The last level at which the match is found (ROOT_LEVEL..leaf_level)
 * \private
*/

int CompactTree::__Search(const Window & window, const unsigned char* p, uint8_t brothers, uint8_t level, TreePath& tree_path) const
{
    int leaf_level = _mask.GetLeafLevel();
	bool in=true;

	while (level <= leaf_level && in==true) {
        uint32_t code=window.TreeHash(level);
        p = __HorizSearch(code,p,brothers,level);
        if (p==NULL)
        {
            in=false;
        }
        else
        {
			//tree_var.ancestor_nodes[level].node=(const unsigned char*)p;
			//tree_var.ancestor_nodes[level]=level;
			tree_path.Push(p);
            brothers=GetNbSons(p);
            p=GetSons(p,level);
            level++;
        };
    };
    return level-1;     // last level in which the word was found
}

/**
 * \brief Search a word in the compact tree, strand \b -, starting at any level higher than root level
 *        Called by __ReverseSearch (1st version)
 * \param [in] window The window encapsulating the word to search
 * \param [in] p The starting node
 * \param [in] The number of brothers (including p)
 * \param [in] level The level in the tree (at least ROOT_LEVEL+1)
 * \param [inout] tree_path The path of the search in the tree
 * \return The last level at which the match is found (ROOT_LEVEL..leaf_level)
 * \private
*/
int CompactTree::__ReverseSearch(const Window & window, const unsigned char* p, uint8_t brothers, uint8_t level, TreePath& tree_path) const
{
    int leaf_level = _mask.GetLeafLevel();
	bool in=true;

	while (level <= leaf_level && in==true) {
        uint32_t code=window.ReverseTreeHash(level);
        p = __HorizSearch(code,p,brothers,level);
        if (p==NULL)
        {
            in=false;
        }
        else
        {
			//tree_var.ancestor_nodes[level].node=(const unsigned char*)p;
			//tree_var.ancestor_nodes[level].level=level;
			tree_path.Push(p);
            brothers=GetNbSons(p);
            p=GetSons(p,level);
            level++;
        };
    };
    return level-1;     // last level in which the word was found
}
/**
 * \brief Search a code in the horizontal branch pointed to by p Called by __Search or __ReverseSearch (2nd version)
 * \param [in] code The code to search for
 * \param [in] p The starting node
 * \param [in] The number of brothers (including p)
 * \param [in] level The level in the tree
 * \return A pointer to the found node, NULL if nothing found
 * \private
*/

const unsigned char* CompactTree::__HorizSearch(uint32_t code, const unsigned char* p, uint8_t brothers, uint8_t level) const
{
    //for(uint8_t i=0; i<brothers && GetCode(p)<code;i++)   // TODO DICHOTOMIE !
    //    p=GetBrother(p,level);

    uint8_t i=0;
    while(i++<brothers)
    {
        if (GetCode(p)<code)
        {
            p=GetBrother(p,level);
            continue;
        };
        if (GetCode(p)==code)
            return p;
        return NULL;
    };
    return NULL;
}

/*=====================================================================
					Retrieving positions in the index tree
=======================================================================*/

/**
 * \brief Walk in the compact tree, retrieving all positions where the seed text is found, 
 *        leaving at the level passed by parameter
 *        If the starting node is one of the ancestors, the hit of tree_var is set accordingly
 *        If p is a leaf, call __RetrievePosTab
 * \param [in] p The seed text to retrieve
 * \param [in] level The level in the tree
 * \param [in] The tree path (used to fix the hit)
 * \param [out] match_pile The match pile used to store the match
 * \private
*/

void CompactTree::__RetrievePos(const unsigned char * p, uint8_t level, TreePath& tree_path, uint32_t qpos, GlintEngineSingleStrand& glintstrandengine, const QueryBuffer& query) const
{
  
	//if ( p == tree_var.ancestor_nodes[level].node ) 
	//	tree_var.Hit=level;    // ancestor_nodes = ancestor Node attribue le veritable hit au match trouve ???  TODO explain better
    tree_path.FixHit(p,level);

	if ( level == ROOT_LEVEL )
	{
        unsigned char sons = GetTopNbSons(p);
        p=GetTopSons(p);
		for (int i=0;i<sons;i++)
		{
			__RetrievePos(p,level+1,tree_path,qpos,glintstrandengine, query);
            p=GetBrother(p,level+1);
		}
	}
	else if ( level == _mask.GetLeafLevel() )
	{
		unsigned char sons = GetNbSons(p);
        const unsigned char *q=GetSons(p,level);
		__RetrievePosTab(q,sons,tree_path.GetHit(),qpos,glintstrandengine, query);
	}
	else
	{
		unsigned char sons = GetNbSons(p);
        p=GetSons(p,level);
		for (int i=0;i<sons;i++)
		{
			__RetrievePos(p,level+1,tree_path,qpos,glintstrandengine, query);
            p=GetBrother(p,level+1);
		}
	}
}

/**
 * \brief Called by __RetrievePos when we are walking at leaf level: for each position, push the match in the match pile
 * \param [in] q The array of positions
 * \param [in] sons The number of positions in q
 * \param [in] The hit of the seed
 * \param [out] match_pile The match pile used to store the match
 * \private
*/
void CompactTree::__RetrievePosTab(const unsigned char * q, uint8_t sons, uint8_t hit, uint32_t qpos, GlintEngineSingleStrand& glintstrandengine, const QueryBuffer& query) const
{
	int j=0;
	uint32_t pos=0;
	const uint32_t *p;
	bool self = _parameters.IsSelf();
	uint64_t mdt = _parameters.GetMaxDistTandem();
	
	p=(uint32_t *)q;
	for (j=0;j<sons;j++)
	{
		pos=*p++;
		if ( !self )
		{
			//glintstrandengine.Push(pos,qpos,tree_var.Hit, query);
			glintstrandengine.Push(pos,qpos,hit,query);
			NbPos++;  //For GlintFreq 
#ifdef DEBUGLOG
			logger->SetLogType("C") << "Adding a new position to the engine " << pos << ' ' << qpos << ' ' << (int) hit << "\n";
#endif
		}
		else if ( pos > qpos && (pos-qpos) < mdt ) {
			//glintstrandengine.Push(pos,qpos,tree_var.Hit, query);
			glintstrandengine.Push(pos,qpos,hit, query);
			NbPos++;  //For GlintFreq 
#ifdef DEBUGLOG
			logger->SetLogType("C") << "Adding a new position to the match pile " << pos << ' ' << qpos << ' ' << (int) hit <<  "\n";
#endif
		}
		else
		{
		}
 
	}
}

/*=====================================================================
   Printing the index (for debugging purpose, used by glintstat)
=======================================================================*/

/** 
 * \brief Print formatted the complete tree, mainly for debugging purposes
 * \param[out] os The output stream
 * \param[in] ct The Compact Tree to print
 * \related CompactTree
*/

ostream & operator<<(ostream& os, const CompactTree& ct) 
{
	os << "MASK = " << ct._mask.GetString() << "\n";
	if ( ct._parameters.IsStatFrequency() == true )
	{
		return ct.__PrintFrequency (os) ;
	}
	else
	{
		return ct.__PrintPositions (os) ;
	}
}

/** 
 * \brief Print formatted the complete tree, mainly for debugging purposes
 * \param[out] os The output stream
*/
ostream& CompactTree::__PrintPositions (ostream& os) const
{
	os << "===================================================\n";
	os << "PRINTING THE POSITIONS: " << nb_top << " HIGH LEVEL BRANCHES\n";
	os << "===================================================\n";

	int nb_of_sons = 0;
	for (uint32_t code=0;code < nb_top;code++) 
	{
        if (GetTopNbSons(code))
        {
			nb_of_sons += GetTopNbSons(code);
            __PrintPhylum(os,code);
        };
	};
	os << "===================================================\n";
	os << "NUMBER OF POSITIONS INDEXED: " << nb_of_sons << "\n";
	os << "===================================================\n";

    return os;
}

/** 
 * \brief Print formatted the words together with their frequency (number of positions)
 * \param[out] os The output stream
 * \param[in] ct The Compact Tree to print
*/
ostream& CompactTree::__PrintFrequency (ostream& os) const
{
	multimap<int,string> freq_word;   // key = number of positions, value = a word

	// Store the words with for each the number of positions found
	int nb_of_sons = 0;
	int nb_of_pos  = 0;
	
    for (uint32_t code=0;code < nb_top;code++) {
        if (GetTopNbSons(code))
        {
			nb_of_sons += GetTopNbSons(code);
            nb_of_pos  += __FreqPhylum(freq_word,code);
        };
	};

	os << "============================\n";
	os << "PRINTING THE WORDS FREQUENCY\n";
	os << "============================\n";
	
	if ( _parameters.IsStatDescending() )
	{
		for ( multimap<int,string>::reverse_iterator i = freq_word.rbegin(); i != freq_word.rend(); ++i )
		{
			os << ToUpper(i->second) << "\t" << i->first << "\n";
		}
	}
	else
	{
		for ( multimap<int,string>::iterator i = freq_word.begin(); i != freq_word.end(); ++i )
		{
			os << ToUpper(i->second) << "\t" << i->first << "\n";
		}
	}		
	os << "===================================================\n";
	os << "NUMBER OF SONS: " << nb_of_sons << "\n";
	os << "NUMBER OF POS : " << nb_of_pos  << "\n";
	os << "===================================================\n";

    return os;
}

/**
 * \brief Keep in freq_word the words and the number of positions
 * \param[out] freq_word[out] A multimap to store the results
 * \param[in] code The code of the phylum to print
 * \private
*/

int CompactTree::__FreqPhylum(multimap<int,string>& freq_word, uint32_t code) const
{
    string w="";
    w = _mask.Reconstruct(w,code,ROOT_LEVEL);
	int stat_level = _parameters.GetStatLevel()!= 0 ? _parameters.GetStatLevel() : _mask.GetLeafLevel();
	int stat_cutoff= _parameters.GetStatCutoff();
    int nb_of_pos = __FreqBranch(freq_word,GetTopSons(code),GetTopNbSons(code),ROOT_LEVEL+1,w);
    if (nb_of_pos >= stat_cutoff && ROOT_LEVEL == stat_level)
    {
		freq_word.insert(pair<int,string>(nb_of_pos,w));
	}
	return nb_of_pos;
}

/**
 * \brief Called by __FreqPhylum, and recursively calling itself
 * \param[out] freq_word[out] A multimap to store the results
 * \param[in] branch A pointer to the starting node
 * \param[in] brothers The number of brothers, including branch
 * \param[in] level The level of the branch in the tree
 * \param[in] reconstructed A partially reconstructed word (will be printed)
 * \private
*/

int CompactTree::__FreqBranch(multimap<int,string>& freq_word, const unsigned char* branch, uint8_t brothers, uint8_t level, const string& word) const
{
	bool is_a_node = (level<_mask.GetLeafLevel());
	int nb_of_pos = 0;
	int stat_level = _parameters.GetStatLevel()!= 0 ? _parameters.GetStatLevel() : _mask.GetLeafLevel();
	int stat_cutoff= _parameters.GetStatCutoff();
    brothers--;     // Do not count myself as a "brother"

    // First, the node
    int code = GetCode(branch);
    string reconstructed = _mask.Reconstruct(word,code,level);

    // Call __FreqBranch on the sons if we are still a node...
    if (is_a_node)
    {
        nb_of_pos += __FreqBranch(freq_word, GetSons(branch,level+1), GetNbSons(branch), level+1, reconstructed);
    }
    else
    {
		nb_of_pos += GetNbSons(branch);
	}
	
	if ( nb_of_pos >= stat_cutoff && level == stat_level )
	{
		freq_word.insert(pair<int,string>(nb_of_pos,reconstructed));
	}

    // Then, the brothers
    const unsigned char* p = branch;
    for (int b=0; b<brothers; b++)
    {
        p = GetBrother(p,level);
        nb_of_pos += __FreqBranch(freq_word,p,1,level,word);
    };
    
    return nb_of_pos;
}

/**
 * * \brief Print formatted the phylum whose code is passed by parameter. 
 *          Called by operator<<
 * \param[out] os The output stream
 * \param[in] code The code of the phylum to print
 * \private
*/

void CompactTree::__PrintPhylum(ostream& os, uint32_t code) const
{
   // os << code << ' ' << Word::Convert(code,Masque->GetLength(ROOT_LEVEL)) << '\n';
    string w="";
    w = _mask.Reconstruct(w,code,ROOT_LEVEL);
    os << w << ' ' << hex << code << " (" << ROOT_LEVEL << '/' << (int)_mask.GetLeafLevel() << ")";
    os << " ==== " << (int) GetTopNbSons(code) << " SONS " << "===================================================\n";
    __PrintBranch(os,GetTopSons(code),GetTopNbSons(code),ROOT_LEVEL+1,w,3);
//    os <<  "Number of sons : " << (int) GetTopNbSons(code) << '\n';
}

/**
 * \brief Print formatted a node, his brothers and his sons. Called by __PrintPhylum
 * \param[out] os The output stream
 * \param[in] branch A pointer to the starting node
 * \param[in] brothers The number of brothers, including branch
 * \param[in] level The level of branch in the tree
 * \param[in] reconstructed A partially reconstructed word (will be printed)
 * \param[in] indent An indentation number
 * \private
*/

void CompactTree::__PrintBranch(ostream& os, const unsigned char* branch, uint8_t brothers, uint8_t level, const string& word, int indent) const
{
	bool is_a_node = (level<_mask.GetLeafLevel());
	int nb_pos = GetNbSons(branch);
	
    brothers--;     // Do not count myself as a "brother"
    string begin="";
    for(int i=0;i<indent;i++) begin+=' ';

    // First, print the node
    int code = GetCode(branch);
    string reconstructed = _mask.Reconstruct(word,code,level);
    os << begin << reconstructed << ' ' << hex << code << " (" << (int)level << '/' << (int)_mask.GetLeafLevel() << ") ";

    // Second, print the son(s) if we are still a node...
	os << dec << (int)brothers << " brothers" << '\n';

    if (is_a_node)
    {
        os << '\n';
        __PrintBranch(os, GetSons(branch,level+1), GetNbSons(branch), level+1, reconstructed, indent+3);
    }
    // ...or print the positions, if we are a leaf
    else
    {
        const unsigned char * pos = GetSons(branch,level);
        os << begin << dec << nb_pos << " positions - ";
        int dsp_pos = (nb_pos<10 || _parameters.IsStatAll()) ? nb_pos : 10;     // Do not display all positions, unless specified
        for (int p=0; p<dsp_pos; p++)
        {
            os << dec << GetPos(pos);
            if ( p<dsp_pos-1)
	            os << ':';
            pos += 4;
        };

        if (dsp_pos==nb_pos)
        {
            os << "\n\n";
        }
        else
        {
            os << "...\n\n";
        };
    };

    // Then, print the brothers
    const unsigned char* p = branch;
    for (int b=0; b<brothers; b++)    //
    {
        p = GetBrother(p,level);
        __PrintBranch(os,p,1,level,word,indent);
    };
}

#undef read_patch

