/*
   $Id: Comparison.cpp 793 2014-08-21 07:32:12Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <cerrno>
#include <vector>
using namespace std;

#ifdef _OPENMP
#include <omp.h>
#endif

#include "Comparison.h"
#include "Display.h"
#include "DisplayM1.h"
#include "DisplaySam.h"
#include "dust.h"
#include "GlintEngine.h"
#include "QueryBufferDouble.h"

#ifdef _OPENMP
/**
 * \brief GlintEngineManager is a manager for Glintengines - used ONLY with openmp
 *        It contains several GlintEngines, associated with QueryBuffers, all cloned from the QueryBuffer passed by parameter
 *        -# When a thread wants to do some job, it calls ReserveEngine
 *        -# When the job is finished, it calls ReleaseEngine
 */

struct GlintEngineStatus {
	GlintEngineStatus(GlintEngineBase* e): eng(e),reserved(false),eof(false) {};
	GlintEngineBase* eng;
	bool reserved;
	bool eof;
};

class GlintEngineManager
{
	public:
		GlintEngineManager(GlintEngineBase*,size_t);
		~GlintEngineManager();
		GlintEngineBase* ReserveEngine( size_t);
		void ReleaseEngine ( size_t );
		bool Eof();
		
	private:
		vector< GlintEngineStatus > __engines;
};

/**
 * \brief We store pairs of GlintEnginePtr (using the cloned QueryBuffer), and a boolean (reserved or not)
 */

GlintEngineManager::GlintEngineManager(GlintEngineBase* eng, size_t nb)
{
	__engines.push_back( GlintEngineStatus(eng) );
	for (size_t i=1; i < nb; ++i)
	{
		__engines.push_back( GlintEngineStatus(eng -> Clone()) );
	}
}

/**
 * \brief delete the cloned engines, (thus the associated QueryBuffers)
 */

GlintEngineManager::~GlintEngineManager()
{
	while (!__engines.size() > 1)
	{
		delete (__engines.back()).eng;
		__engines.pop_back();
	}
}

/**
 * \brief Return true if and only if all engines have eof flag set
 */

bool GlintEngineManager::Eof()
{
	for ( size_t i=0; i<__engines.size(); ++i )
	{
		if ( !__engines[i].eof )
		{
			return false;
		}
	}
	return true;
}

/**
 * \brief ReserveEngine: If the Engine is free, call ReadNextSeq() and return the engine ReadNextSeq returned true
 *                       If ReadNextSeq() returned false, ReserveEngine returns NULL and sets the eof flag
 * 
 * \param tid The thread id
 * \return The engine, or NULL
 */
GlintEngineBase* GlintEngineManager::ReserveEngine( size_t tid )
{
	GlintEngineBase* rvl = NULL;
	#pragma omp critical
	{
		if ( __engines[tid].reserved )
		{
			throw(logic_error((string) "ERROR - This engine is already reserved ! " + ERROR_FILE_LINE(__FILE__,__LINE__)));
		}
		else
		{
			if (__engines[tid].eng -> ReadNextSeq())
			{
				__engines[tid].reserved = true;
				rvl =  __engines[tid].eng;
			}
			else
			{
				__engines[tid].eof = true;
			}
		}
	}
	return rvl;
}
	
/**
 * \brief ReleaseEngine: If the engine is reserved, mark it as free
 * \param tid The thread id
 */

void GlintEngineManager::ReleaseEngine ( size_t tid )
{
	#pragma omp critical
	{
		if (__engines[tid].reserved)
		{
			__engines[tid].reserved = false;
		}
		else
		{
			throw(logic_error((string) "ERROR - No engine to release here !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		}
	}
}
#endif

/** 
 * \brief Initializing the Display object: we create an object derived of Display
 *        - The Display type created depends on the -M switch
 * \return dsp_ptr An autoptr to the new display
 * \exception Invalid argument
*/

auto_ptr<Display> ComparisonBase::_InitDisplay()
{
    auto_ptr<Display> dsp_ptr(NULL);
    switch ( _parameters.GetOutputFormat() ) 
    {
		case OUTPUT_FORMAT_M0   : dsp_ptr.reset(new DisplayM0()); break;
		case OUTPUT_FORMAT_M1   : dsp_ptr.reset(new DisplayM1(CompactTree::Instance().GetSeq())); break;
		case OUTPUT_FORMAT_M2   : dsp_ptr.reset(new DisplayM2()); break;
		case OUTPUT_FORMAT_M3   : dsp_ptr.reset(new DisplayM3()); break;
		case OUTPUT_FORMAT_M4   : dsp_ptr.reset(new DisplayM4()); break;
		case OUTPUT_FORMAT_M5   : dsp_ptr.reset(new DisplayM5()); break;
		case OUTPUT_FORMAT_M8   : dsp_ptr.reset(new DisplayM8()); break;
		case OUTPUT_FORMAT_M9   : dsp_ptr.reset(new DisplayM9()); break;
		case OUTPUT_FORMAT_FA   : dsp_ptr.reset(new DisplayFasta()); break;
		case OUTPUT_FORMAT_FASTQ: dsp_ptr.reset(new DisplayFastq()); break;
		case OUTPUT_FORMAT_SAM  : dsp_ptr.reset(new DisplaySAM()); break;
		case OUTPUT_FORMAT_BAM  : dsp_ptr.reset(new DisplayBAM()); break;
		case OUTPUT_FORMAT_SAMEXT : dsp_ptr.reset(new DisplaySAM(true)); break;
		case OUTPUT_FORMAT_BAMEXT : dsp_ptr.reset(new DisplayBAM(true)); break;		
		case OUTPUT_FORMAT_BED  : dsp_ptr.reset(new DisplayBed()); break;
		default: throw (invalid_argument("ERROR - Unknown output format"));
	}
    dsp_ptr->DisplayTopHeader();
    return dsp_ptr;
}

/**
 *  \brief Start the comparison process (2 versions, for OPENMP and for sequential code)
 *         -# Create, use and destroy an object deriving from QueryBuffer
 *         -# Create, use and destroy an object deriving from Display
 *         -# Create, use and destroy an object deriving from GlintEngineBase
 *         -# while there are sequences to read, call _Run
 * 
 */

void ComparisonBase::Run()
{
	// Create the GlintEngineManager
	QueryBuffer* master_query_ptr = _InitQueryBuffer();
	if ( _parameters.IsToUpper() ) QueryBuffer::ToUpper( true );

	// Using an auto_ptr for display, because there is only ONE display object
	auto_ptr<Display>  display_aptr (_InitDisplay());
	auto_ptr<GlintEngineBase> master_engine_aptr(NULL);
	if (_parameters.IsPairEnd() || _parameters.IsePCR())
	{
		master_engine_aptr.reset(_InitGlintEngine((QueryBufferDouble*)master_query_ptr,*display_aptr));
	}
	else
	{
		master_engine_aptr.reset(_InitGlintEngine(master_query_ptr,*display_aptr) );
	}
	
	if ( !_parameters.IsSilent() )
		cerr << "Processing the request\n";
#ifdef _OPENMP
	GlintEngineManager engine_mngr(master_engine_aptr.get(),omp_get_max_threads());
	//GlintEngineManager engine_mngr(master_engine_aptr.get(),2);
	
	while ( !engine_mngr.Eof() )
	{
		#pragma omp parallel shared(engine_mngr)
		{	
			#pragma omp for
			for ( int i = 0; i<10000; i++ )
			////for ( int i = 0; i<20; i++ )
			{
				int tid = omp_get_thread_num();
				GlintEngineBase* engine_ptr = NULL;
				engine_ptr = engine_mngr.ReserveEngine(tid);
				if ( engine_ptr != NULL )
				{
					engine_ptr -> Run();
					engine_mngr.ReleaseEngine(tid);
				}
			}
		}
	}
#else
 	while ( master_query_ptr -> ReadNextSeq() )
	{
		master_engine_aptr -> Run();
	}
#endif

	if ( !_parameters.IsSilent() )
		cerr <<"Finished\n";
		
}

/**
 * \brief Constructing the correct GlintEngine according to the input format
 * \param q_ptr, an already built QueryBuffer*
 * \pre   The _parameters singleton must be initialized
 * \public
 */

QueryBuffer* Comparison::_InitQueryBuffer() 
{ 
	string query_filename = _parameters.GetQueryFilename().GetName();
	string bwd_filename   = _parameters.GetBwdFilename().GetName();
	CompressionAlgorithm compression = _parameters.GetQueryFilename().GetCompression();
	
	bool quality = (_parameters.GetOutputFormat()==OUTPUT_FORMAT_BAM || 
					_parameters.GetOutputFormat()==OUTPUT_FORMAT_BAMEXT || 
					_parameters.GetOutputFormat()==OUTPUT_FORMAT_SAM || 
					_parameters.GetOutputFormat()==OUTPUT_FORMAT_FASTQ ); 
	bool pair_end = ( _parameters.IsPairEnd() || _parameters.IsePCR() );
	

	QueryBuffer* q = NULL;
	if ( !pair_end )
	{
		switch ( _parameters.GetInputFormat() )
		{
			case INPUT_FORMAT_FA   : q = new QueryBufferFasta(query_filename,compression); break;
			case INPUT_FORMAT_FASTQ: q = new QueryBufferFastq(query_filename,compression,quality); break;
			case INPUT_FORMAT_SAM  : q = new QueryBufferSam  (query_filename,false,quality); break;
			case INPUT_FORMAT_BAM  : q = new QueryBufferSam  (query_filename,true, quality); break;
			default                : throw (runtime_error( "ERROR - INPUT FORMAT NOT SUPPORTED" ) );
		}
	}
	else
	{
		switch ( _parameters.GetInputFormat() )
		{
			case INPUT_FORMAT_BAM : q  = new QueryBufferSamDouble(query_filename,quality); break;
			case INPUT_FORMAT_FA  : q  = new QueryBufferFastaDouble(query_filename, bwd_filename, compression,false); break;
			case INPUT_FORMAT_FASTQ: q = new QueryBufferFastqDouble(query_filename, bwd_filename, compression,quality); break;
			default               : throw (runtime_error( "ERROR - INPUT FORMAT NOT SUPPORTED IN --pair-end" ) );
		}
	}		
	return q;
}

/**
 *  \brief The constructor
 *  \pre   The _parameters singleton must be initialized
 */

ComparisonBase::ComparisonBase():
	_parameters(Parameters::Instance()),
	_ctree (CompactTree::Instance()),
	__mask(Mask::Instance())
{}

/**
 *  \brief validate the query file. Called by the constructor
 *         - Check the query file (reading only the first two lines), to be sure it is a fasta file
 *  \public
 */
bool ComparisonBase::__ValidQueryFile()
{
	bool valid = true;
	string filename                  = _parameters.GetQueryFilename().GetName();
	CompressionAlgorithm compression = _parameters.GetQueryFilename().GetCompression();

	if ( filename=="-" ) // Checking stdin as a valid fasta file is not implemented yet
	{
		return valid;
	}

	IstreamPtr infile_ptr ( filename, "", compression );
	istream & infile = *infile_ptr;
	string s;
	
	 getline(infile, s); //The first line
	 if ( s[0] != '>' )
		valid = false;
	 getline(infile, s); //The second line
	 if (s.length() == 0 )
		valid = false;

	return valid;
}

/**
 *  \brief Start the comparison process
 *         -# Create a query and a match_results
 *         -# call _Run
 */
void ComparisonSelf::Run()
{
	QueryBuffer*  query_ptr  = _InitQueryBuffer();
	if ( query_ptr -> ReadNextSeq() == false )
	{
		throw ( runtime_error("ERROR - No sequence to read !"));
	}

	auto_ptr<Display> display_aptr = _InitDisplay();
	auto_ptr<GlintEngineBase> engine_aptr(_InitGlintEngine(query_ptr,*display_aptr));
	
	if ( !_parameters.IsSilent() )
		cerr << "Processing the request\n";

	engine_aptr -> Run();
	
	if ( !_parameters.IsSilent() )
		fprintf(stderr,"Finished\n");
	
}
