/*
   $Id: SuitableRegion.cpp 680 2012-01-30 14:42:53Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <cerrno>
#include <vector>
#include <limits>
using namespace std;

#ifdef _OPENMP
#include <omp.h>
#endif

#include "SuitableRegion.h"

SuitableRegion::SuitableRegion():
	_parameters(Parameters::Instance()),
	_query_ptr(_InitQueryBuffer()),
	_display(DisplayFreq())
{
	
	__SentenceScore[ LOW ] = -_parameters.GetConseqFreq()/(100 - _parameters.GetConseqFreq());
	__SentenceScore[ HIGH ] = 1;
	
	__SuitableScore[ LOW ] = 1;
	__SuitableScore[ HIGH ] = -_parameters.GetConseqUnFreq()/(100 - _parameters.GetConseqUnFreq());
	__SuitableScore[ CONDEMNED ] = -numeric_limits<double>::max();	
}

/**
 * \brief Start the comparison between the sequence stored in the QueryBuffer and the bank singleton
 *         -# Mask low complexity sequences if asked ( -C > 0)
 *         -# Search for seeds on the plus and on the minus strands
 *         -# Align the seeds and create the hsps on the plus and on the minus strands
 */
void SuitableRegion::Run ()
{		
	while ( _query_ptr -> ReadNextSeq() )
	{
		__MaskRepeats();	
		__DetectSuitableRegions();
		
		_display.SetQuery( _GetQueryPtr() );	
		_display.DisplaySuitableFasta( _GetFrequentBuffer(), _GetSuitableRegions() );
	}
}

/**
 * \brief Recode the sequence according to the soft masking
 *         -# Foreach window (length of mask_len) count the occurence of the word in the bank index
 *             - if count > CountCutoff of word in black list code = HIGH
 *             - else code = LOW
 *         -# The recoded sequence is stored in the frequentCodeBuffer buffer
 */
void SuitableRegion:: __MaskRepeats()
{
	size_t              seq_length = _query_ptr -> GetSeqLength();
	const unsigned char * seq_base = _query_ptr -> GetBase();
	
	freqCode f;	
	_ClearFrequentBuffer();

	// The pos variable corresponds to the first base of the sequence (absolute position)
	for (uint32_t pos=1; pos <= seq_length; pos++)
	{
		char c=*(seq_base + pos);
		bool IsFrequent = ( IsLower(c) || c == 'N' ) ? true : false;			
		f = (IsFrequent) ? HIGH : LOW;
		_FrequentBufferPut( f );	
	}
		
}

/**
 * \brief Detect the suitable regions
 *         -# Regions with an excess of frequent words are sentenced
 *         -# Regions with a sufficient proportion of unfrequent words are selected
 */
void SuitableRegion::__DetectSuitableRegions()
{
	__SentencingRegions(); 	
	__SelectingSuitableRegions();	
}

/** \brief Identify condemned zone using a scoring system
 *        -#  Large regions with an excess of frequent word are sentenced 
 *             More precisely, we are looking here for the smallest regions satisfying the constraints (  size>CondemnedMinSize with frequent word > ConseqFreq )
 *     How the algortihm works :
 *            S[i]   :  dynamic programming score :   S(i) = max(  0, S(i-1) + s(i) )
 *            S[i]>0 :  the proportion of high frequency words is > ConseqFreq
 *   Le f denote maximum accepted frequency of frequent words : 
 *       n1 : number of frequent words
 *       n2 : number of unfrequent words     
 *   n1/(n1+n2) > 0 <=> S[i] > 0 with S[i] = n1+k*n2 with k = - f/(1-f)
 *           
*/
void SuitableRegion::__SentencingRegions() 
{		
	uint32_t  size  = _GetBufferSize();
	freqCode* base  = _GetBufferBase();
	uint32_t      CondemnedMinSize = _parameters.GetCondemnedMinSize();
	
	__sentencedRegions.clear();
	
	uint32_t current=0, previous=0;	
	uint32_t start = 0, rlength = 0;
	
	for ( uint32_t i = 1  ; i <= size ; i++, base++ )
	{ 
		current = __max0( previous + __SentenceScore[ *base ] );		
		rlength = i - start;	
		if ( current > 0 && rlength > CondemnedMinSize)
		{	
			regionFreq r = { start + 1 , i  };
			__sentencedRegions.push_back(r);
			current = 0;	
		}					
		if ( current == 0 ) 
			start = i;
		previous = current;	
	}	
	__MaskSentencedRegions();			
}

/** \brief Identify suitable regions using a scoring system
 *        -#  Large regions with an excess of unfrequent word are selected
 *             More precisely, we are looking here for the largest regions satisfying the constraints (  size>SuitableMinSize with unfrequent word > ConseqUnFreq )
 *     How the algortihm works :
 *            S[i]   :  dynamic programming score :   S(i) = max(  0, S(i-1) + s(i) )
 *            S[i]>0 :  the proportion of low frequency words is > ConseqUnFreq
 *                
*/
void SuitableRegion::__SelectingSuitableRegions() 
{		
	uint32_t   size   = _GetBufferSize();
	freqCode*  base  = _GetBufferBase();
	uint32_t   SuitableMinSize  = _parameters.GetSuitableMinSize();
	uint32_t   SuitableMaxSize  = _parameters.GetSuitableMaxSize();
	/*uint8_t msk_len = _mask.GetLength();
	uint8_t msk_offset = msk_len-1;*/
	
	uint32_t current=0, previous=0;	
	uint32_t start = 0, rlength = 0;
	
	for ( uint32_t i = 1  ; i <= size ; i++, base++ )
	{ 		
		current = __max0( previous + __SuitableScore[ *base ] );		
		rlength = i - start + 1;		
			
		if ( current == 0 && previous > 0 && rlength > SuitableMinSize )
		{	
			__AddSuitableRegion( start + 1 , i - 1  );
			/*regionFreq r = { start + 1 , i - 1 + msk_offset};
			__suitableRegions.push_back(r);	*/
		}	
		if ( current > 0 && rlength > SuitableMaxSize )
		{
			__AddSuitableRegion( start + 1 , i );
			current = 0;	
		}		
		if ( current == 0 ) 
			start = i;	
		previous = current;	
	}
	if ( current > 0 && rlength > SuitableMinSize )
	{
		__AddSuitableRegion( start + 1 , size  );
		/*regionFreq r = { start + 1 , size + msk_offset};
		__suitableRegions.push_back(r);*/
	}	
	
	cerr<<__suitableRegions.size()<<endl;
	
}

/** \brief Splitting suitable regions
 *        -#  Large regions (> MAX_SUITABLE_SIZE) are spitted in smaller subregions
 *             
*/
void SuitableRegion::__AddSuitableRegion( uint32_t start, uint32_t end )
{
	uint32_t SuitableMinSize   = _parameters.GetSuitableMinSize();
	uint32_t SuitableChunkSize = _parameters.GetSuitableChunkSize();
	uint32_t size = end - start + 1;
		
	if ( size <= _parameters.GetSuitableMaxSize() )
	{
		regionFreq r = { start, end };

		uint32_t r_size = r.end - r.start + 1;
		if ( r_size >= SuitableMinSize )
			__suitableRegions.push_back(r);	

	}
	else
	{
		uint32_t r_start = start;
		uint32_t r_end = 0;
		while ( r_end != end )
		{	
			r_end = ( r_start + SuitableChunkSize -1 <= end ) ? r_start + SuitableChunkSize -1 : end;
			regionFreq r = { r_start, r_end };

			uint32_t r_size = r.end - r.start + 1;
			if ( r_size  >= SuitableMinSize )
			 __suitableRegions.push_back(r);	

			r_start = r_end + 1;
		}		
	}
}

/** \brief Mask sentenced regions by C
 *              
*/
void SuitableRegion:: __MaskSentencedRegions() 
{
	freqCode * fbase = _GetBufferBase();
	
	for ( vector<regionFreq>::iterator rit = __sentencedRegions.begin(); rit != __sentencedRegions.end(); ++rit )
	{ 
		for( uint32_t i = rit->start ; i <= rit->end ; ++i )
		{	
			fbase[i-1] = CONDEMNED;   //O-based buffer
		}
	}		
}

/**
 * \brief displays the frequenCode Buffer in fasta format (for debugging)
 */
void SuitableRegion::_DisplayFrequentBuffer()
{	
	uint32_t   size  = _GetBufferSize();
	freqCode* fbase  = _GetBufferBase();
	
	unsigned char base;
	for ( uint32_t i = 0  ; i < size ; ++i )
	{ 
		base = *(fbase+i);	
		if ( i && i%60==0 )
			cout<<endl;
		cout<<Decode(base);
	}
	if ( size!=0 )	
		cout<<endl;
}






